/*
 * Copyright (c) 2012 Tommy Hallgren
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * Coding principles:
 * - The framework uses the "Factory constructor pattern" since is seems as less "magic", less prone
 *   to "this" mix-ups. Read more here: http://javascript.info/tutorial/oop
 *
 * - Name the variable that acts as both this and base to the name of the class,
 *   but with a lower-case initial letter.
 *
 * - Try to exit from functions as soon as an answer is known.
 */
var Widjet = {};
Widjet.ui = {};
Widjet.behavior = {};
Widjet.util = {};
Widjet.extras = {};
Widjet.mvp = {};

Widjet.init = function () {
    // No framework initialization needed so far.
};

Widjet.uniqueIdCount = 0;
Widjet.idPrefix = "_";

Widjet.getUniqueId = function () {
    return Widjet.idPrefix + Widjet.uniqueIdCount++;
};

//
// The core classes in this file are:
// - Widjet.ui.Widget
// - Widjet.ui.Element
// - Widjet.ui.Panel
//
// The rest of the Widjet.ui classes are convenience classes mostly that makes your widgets talk about
// TextBoxes and PasswordBoxes instead of InputElement("text") and InputElement("password");
//

// Widget acts as a base class and has functions for nestling Widjet.
// It has no corresponding type in the DOM.
Widjet.ui.Widget = function (domNode) {
    var widget = {};

    Widjet.behavior.DomNode(widget);
    widget.domNode(domNode);

    Widjet.behavior.Nestable(widget);

    return widget;
};

// Element is a DOM Element Node
Widjet.ui.Element = function (tag, className) {
    var element = Widjet.ui.Widget(document.createElement(tag));

    element.className = function (className) {
        if (arguments.length !== 1)
            return element.domNode().className;
        element.domNode().className = className;
    };

    element.addClassName = function (classNameToAdd) {
        element.domNode().classList.add(classNameToAdd);
    };

    element.removeClassName = function (classNameToRemove) {
        element.domNode().classList.remove(classNameToRemove);
    };

    element.text = function (text) {
        return Widjet.util.textContent(element.domNode(), text);
    };

    element.value = function (value) {
        if (arguments.length !== 1)
            return element.domNode().value;
        element.domNode().value = value;
    };

    element.html = function (markup) {
        if (arguments.length !== 1)
            return element.domNode().innerHTML;
        element.domNode().innerHTML = markup;
    };

    element.disabled = function (disabled) {
        if (arguments.length !== 1)
            return element.domNode().getAttribute("disabled");

        if (disabled || disabled === "disabled")
            element.domNode().setAttribute("disabled", "disabled");
        else
            element.domNode().removeAttribute("disabled");
    };

    element.position = function (position) {
        if (arguments.length !== 1)
            return Widjet.util.getNodePosition(element.domNode());
        Widjet.util.setNodePosition(element.domNode(), position)
    };

    Widjet.behavior.DomEvents(element);
    Widjet.behavior.CustomEvents(element);

    element.className(className);

    return element;
};

// Panel is a <tag> element with support for adding other widgets as children
// Example, a panel could be a <ul> and each widget inside could be <li>
// Or a <select> panel with <option> or <optgroup> inside.
Widjet.ui.Panel = function (tag, className) {
    var panel = Widjet.ui.Element(tag, className);

    Widjet.behavior.Container(panel);

    return panel;
};

// Text is a DOM Text Node
Widjet.ui.Text = function (newText) {
    var text = Widjet.ui.Widget(document.createTextNode(text));

    text.text = function (newText) {
        if (newText === undefined)
            return text.domNode().textContent;

        text.domNode().textContent = newText;
    };

    text.text(newText);
    return text;
};

// Literal is short for Element("span") with the text inside
Widjet.ui.Literal = function (text, className) {
    var literal = Widjet.ui.Element("span", className);
    literal.text(text);
    return literal;
};

// Button is a <button>
Widjet.ui.Button = function (text, className) {
    var button = Widjet.ui.Element("button", className);
    button.domNode().setAttribute("type", "button");
    button.text(text);
    return button;
};

// Label is a <label> element
Widjet.ui.Label = function (text, formElement, className) {
    var label = Widjet.ui.Element("label", className);

    label.text(text);
    label.domNode().setAttribute("for", formElement.id());

    return label;
};

// FormElement is an Element with typical input control functionality.
Widjet.ui.FormElement = function (tag, className) {
    var formElement = Widjet.ui.Element(tag, className);

    formElement.focus = function () {
        if (formElement.domNode().focus !== undefined)
            formElement.domNode().focus();
    };

    return formElement;
};

// TextArea is a <textarea> FormElement.
Widjet.ui.TextArea = function (className) {
    return Widjet.ui.FormElement("textarea", className);
};

// InputElement is an <input> of specified type
Widjet.ui.InputElement = function (type, className) {
    var inputElement = Widjet.ui.FormElement("input", className);

    inputElement.domNode().setAttribute("type", type);

    return inputElement;
};

// TextBox is an <input type="text">
Widjet.ui.TextBox = function (className) {
    return Widjet.ui.InputElement("text", className);
};

// PasswordBox is an <input type="password">
Widjet.ui.PasswordBox = function (className) {
    return Widjet.ui.InputElement("password", className);
};

// CheckBox is an <input type="checkbox">
Widjet.ui.CheckBox = function (className) {
    var checkbox = Widjet.ui.InputElement("checkbox", className);

    checkbox.checked = function (checked) {
        if (checked === undefined)
            return checkbox.domNode().checked;
        checkbox.domNode().checked = checked;
    };

    return checkbox;
};

// Slider is an <input type="range">
Widjet.ui.Slider = function (className) {
    var slider = Widjet.ui.InputElement("range", className);

    slider.min = function (min) {
        if (min === undefined)
            return slider.domNode().getAttribute("min");
        slider.domNode().setAttribute("min", min);
    };

    slider.max = function (max) {
        if (max === undefined)
            return slider.domNode().getAttribute("max");
        slider.domNode().setAttribute("max", max);
    };

    slider.step = function (step) {
        if (step === undefined)
            return slider.domNode().getAttribute("step");
        slider.domNode().setAttribute("step", step);
    };

    slider.value = function (value) {
        if (value === undefined)
            return slider.domNode().value;
        slider.domNode().value = value;
    };

    return slider;
};

// RadioBox is an <input type="radio">
Widjet.ui.RadioBox = function (className) {
    // TODO: Make an example that uses this
    // TODO: Support configuring groups of mutually exclusiveness on a radio button
    var inputElement = Widjet.ui.InputElement("radio", className);

    inputElement.checked = function (checked) {
        if (checked === undefined)
            return inputElement.domNode().checked;
        inputElement.domNode().checked = checked;
    };

    return inputElement;
};

// FieldSet is a <fieldset> with a nested <legend>
Widjet.ui.FieldSet = function (text, className) {
    var legend = Widjet.ui.Element("legend");
    legend.text(text);

    var panel = Widjet.ui.Panel("fieldset", className);
    panel.addWidget(legend);
    return panel;
};


//
// Widjet.behavior.*
//

//
// Widjet.behavior.DomNode
// - Makes a widget have a domNode() property and an id() property
//
// Methods:
// - domNode(newDomNode) - gets or sets the domNode property
// - id() - gets the ID of the attached domNode. An id is generated if it don't exist already
// - adoptNodeById(id) - find the domNode by given id and sets the domNode property
//
Widjet.behavior.DomNode = function (widget) {
    var domNode = undefined;

    widget.domNode = function (newDomNode) {
        if (arguments.length !== 1)
            return domNode;
        domNode = newDomNode;
    };

    widget.id = function () {
        if (!domNode.id)
            domNode.id = Widjet.getUniqueId();

        return domNode.id;
    };

    // Warning!
    // Only call adoptNode*() before calls to:
    // - id()
    // - and any event bindings
    widget.adoptNodeById = function (id) {
        widget.domNode(document.getElementById(id));
    };
};

//
// Widjet.behavior.DomEvents
// - Makes the widget capable of utilizing DOM events
//
// Methods:
// - bindEvent(eventName, fn)
// - unbindEvent(eventName, fn)
//
// Requires the widget to have a domNode() function
//
Widjet.behavior.DomEvents = function (widget) {
    // bind/unbindEvent are just convenience wrappers for normal DOM events
    widget.bindEvent = function (eventName, fn) {
        Widjet.util.bindEvent(widget.domNode(), eventName, fn);
    };

    widget.unbindEvent = function (eventName, fn) {
        Widjet.util.unbindEvent(widget.domNode(), eventName, fn);
    };
};

//
// Widjet.behavior.Nestable
// - Makes a widget nestable by having a parent
//
// Methods:
// - parent(newParent) - gets or sets the parent widget
//
Widjet.behavior.Nestable = function (widget) {
    var parent = undefined;

    widget.parent = function (newParent) {
        if (arguments.length !== 1)
            return parent;

        if (parent !== undefined && newParent !== undefined)
            throw "Widget already has a parent, failed to set parent";

        parent = newParent;
    };
};

//
// Widjet.behavior.Container
// - Adds behavior to make a widget a container of other widgets. Requires that widget is Nestable
//
// Methods:
// - addWidget(widget) - adds the widget as a child of the container
// - removeWidget(widget) - removes the widget as a child of the container
// - clear() - removes all children from the container
//
Widjet.behavior.Container = function (container) {
    container.widgets = [];

    container.addWidget = function (widget) {
        widget.parent(container);
        container.widgets.push(widget);
        container.domNode().appendChild(widget.domNode());
    };

    container.removeWidget = function (widget) {
        container.domNode().removeChild(widget.domNode());
        widget.parent(undefined);
        container.widgets = Widjet.util.filter(container.widgets, function (x) {
            return x !== widget;
        });
    };

    container.clear = function () {
        Widjet.util.forEach(container.widgets, function (widget) {
            widget.parent(undefined);
        });
        container.widgets = [];
    };
};

//
// Widjet.behavior.CustomEvents
// - Adds behavior for a widget to host non-DOM events, typically more logic oriented events
//
// Methods:
// - supportsCustomEvent(eventName) - returns true if the widget can fire eventName custom events
// - addCustomEvent(eventName) - registers an eventName so it can be listened upon
// - triggerCustomEvent(eventName, arg) - fires eventName custom event and send arg to all bound functions
// - bindCustomEvent(eventName, fn) - fn will be called when eventName is fired
// - unbindCustomEvent(eventName, fn) - unbinds fn to be called when eventName is fired
Widjet.behavior.CustomEvents = function (widget) {
    var customEvents = {};

    widget.supportsCustomEvent = function (eventName) {
        return customEvents.hasOwnProperty(eventName);
    };

    widget.addCustomEvent = function (eventName) {
        if (customEvents[eventName] !== undefined)
            return;

        customEvents[eventName] = [];
    };

    // Custom events must be triggered manually by code
    widget.triggerCustomEvent = function (eventName, arg) {
        var boundFunctions = customEvents[eventName];
        if (boundFunctions === undefined)
            return;

        Widjet.util.forEach(boundFunctions, function (fn) {
            fn(arg);
        });
    };

    widget.bindCustomEvent = function (eventName, fn) {
        // A widget has to call addCustomEvent in its constructor
        // for others to be able to listen to that event
        if (!customEvents.hasOwnProperty(eventName))
            throw "No custom event with name " + eventName + " has been added";

        // Check if "fn" already has been registered for "eventName"
        var boundFunctions = customEvents[eventName];
        for (var i = 0; i < boundFunctions.length; i++) {
            if (boundFunctions[i] === fn) {
                throw "The function has already been bound to the event " + eventName;
            }
        }

        boundFunctions.push(fn);
    };

    widget.unbindCustomEvent = function (eventName, fn) {
        if (!customEvents.hasOwnProperty(eventName))
            throw "No custom event with name " + eventName + " has been added";

        var boundFunctions = customEvents[eventName];
        for (var i = 0; i < boundFunctions.length; i++) {
            if (boundFunctions[i] === fn) {
                boundFunctions.splice(i, 1);
                return;
            }
        }
        throw "The function has not been bound to the event " + eventName;
    };
};

//
// Widjet.util.*
//

Widjet.util.textContent = function (domNode, text) {
    if (text === undefined) {
        if (typeof domNode.textContent !== "undefined")
            return domNode.textContent;

        if (typeof domNode.innerText !== "undefined")
            return domNode.innerText;

        throw "Browser does not support textContent nor innerText";
    }

    if (typeof domNode.textContent !== "undefined") {
        domNode.textContent = text;
        return;
    }

    if (typeof domNode.innerText !== "undefined") {
        domNode.innerText = text;
        return;
    }

    throw "Browser does not support textContent nor innerText";
};

Widjet.util.bindEvent = function (domNode, eventName, fn) {
    if (domNode.addEventListener) {
        domNode.addEventListener(eventName, fn);
        return;
    }

    if (domNode.attachEvent) {
        domNode.attachEvent("on" + eventName, fn);
        return;
    }

    throw "This browser does not support addEventListener() nor attachEvent()";
};

Widjet.util.unbindEvent = function (domNode, eventName, fn) {
    if (domNode.removeEventListener) {
        domNode.removeEventListener(eventName, fn);
        return;
    }

    if (domNode.detachEvent) {
        domNode.detachEvent("on" + eventName, fn);
        return;
    }

    throw "This browser does not support removeEventListener() nor detachEvent()";
};

Widjet.util.property = function (obj, propertyName) {
    return function (value) {
        var callable = typeof(obj[propertyName]) === "function";

        if (arguments.length === 1)
            setValue(value);
        else
            getValue();

        function setValue(value) {
            if (callable)
                obj[propertyName](value);
            else
                obj[propertyName] = value;
        }

        function getValue() {
            // Return the value of the property
            if (callable)
                return obj[propertyName]();
            return obj[propertyName];
        }
    };
};

Widjet.util.parseXML = function (xml) {
    if (DOMParser)
        return new DOMParser().parseFromString(xml, "text/xml");

    if (ActiveXObject) {
        var parser = new ActiveXObject("Microsoft.XMLDOM");
        parser.async = "false";
        return parser.loadXML(xml);
    }

    throw "No XML parser found";
};

Widjet.util.requestXml = function (url, successFunction, errorFunction) {
    var request = new XMLHttpRequest();
    request.open('GET', url, true);
    request.onreadystatechange = handleStateChange;
    request.send();

    function handleStateChange() {
        if (request.readyState !== 4)
            return;

        if (request.status === 200) {
            if (successFunction)
                successFunction(request);
            return;
        }

        if (errorFunction)
            errorFunction(request);
    }
};

Widjet.util.benchmark = function (fn, times) {
    times = times || 10000;

    var n = +new Date;
    for (var i = 0; i < times; i++)
        fn();
    return "" + (+new Date - n) + 'ms';
};

Widjet.util.forEach = function (arr, fn) {
    for (var i = 0; i < arr.length; i++)
        fn(arr[i]);
};

// Returns a new collection with the items that passed the given predicate
Widjet.util.filter = function (arr, predicate) {
    var result = [];

    for (var i = 0; i < arr.length; i++) {
        if (predicate(arr[i]))
            result.push(arr[i]);
    }

    return result;
};

// Returns true if any of the items in arr passes the given predicate
Widjet.util.any = function (arr, predicate) {
    for (var i = 0; i < arr.length; i++) {
        if (predicate(arr[i]))
            return true;
    }

    return false;
};

// Returns true if all of the items in arr passes the given predicate
Widjet.util.all = function (arr, predicate) {
    for (var i = 0; i < arr.length; i++) {
        if (!predicate(arr[i]))
            return false;
    }

    return true;
};

Widjet.util.getNodePosition = function (node) {
    // Thank you - http://www.quirksmode.org/js/findpos.html
    // TODO: Examine if this really gives the right coordinates, draggables seem to "jump" at first mouse movement
    var x = 0;
    var y = 0;
    do {
        if (node.nodeType === 1) {
            x += node.offsetLeft;
            y += node.offsetTop;
        }
        node = node.offsetParent;
    } while (node);
    return { x:x, y:y };
};

Widjet.util.setNodePosition = function (node, position) {
    node.style.left = "" + position.x + "px";
    node.style.top = "" + position.y + "px";
};


//
// Widjet.extras.*
//

// Draggable is a panel that can be dragged around with the mouse
Widjet.extras.Draggable = function (tag, className) {
    var panel = Widjet.ui.Panel(tag, className);

    var diffX = 0;
    var diffY = 0;

    panel.bindEvent("mousedown", function (event) {
        var pos = panel.position();

        diffX = pos.x - event.clientX;
        diffY = pos.y - event.clientY;

        Widjet.util.bindEvent(document, "mousemove", docMouseMove);
        Widjet.util.bindEvent(document, "mouseup", stopDragging);
    });

    return panel;

    function docMouseMove(ev) {
        var x = ev.clientX + diffX;
        var y = ev.clientY + diffY;
        panel.position({ x:x, y:y });
    }

    function stopDragging() {
        Widjet.util.unbindEvent(document, "mousemove", docMouseMove);
        Widjet.util.unbindEvent(document, "mouseup", stopDragging);
    }
};

//
// Widjet.extras.TabStrip
// TabStrip is a <ul> with <li> for each tab added.
//
// Markup used:
// <ul class="the className you provide in the constructor">
//     <li class="active">The selected and highlighted tab</li>
//     <li>Unselected tab</li>
//     <li>Also an unselected tab</li>
// </ul>
//
// Events:
// - customEvent "tabClicked", argument is the tabId the user clicked on
//
Widjet.extras.TabStrip = function (className) {
    var activeTabId = "";
    var tabs = {};

    var tabStrip = Widjet.ui.Panel("ul", className);

    tabStrip.addCustomEvent("tabClicked");

    tabStrip.addTab = function (text, tabId) {
        if (tabs[tabId] !== undefined)
            throw "Tab with tabId " + tabId + " has already been added";

        var tab = Widjet.ui.Panel("li");
        tabs[tabId] = tab;
        tab.text(text);
        tabStrip.addWidget(tab);
        tab.bindEvent("click", tabClick(tabId));
    };

    tabStrip.removeTab = function (tabId) {
        if (activeTabId === tabId)
            activeTabId = "";

        if (!tabs.hasOwnProperty(tabId))
            throw "Tab with tabId " + tabId + " does not exist";

        var tab = tabs[tabId];
        tabStrip.removeWidget(tab);
        delete tabs[tabId];
    };

    tabStrip.activeTab = function (tabId) {
        if (tabId === undefined)
            return activeTabId;

        if (activeTabId !== "")
            tabs[activeTabId].removeClassName("active");

        tabs[tabId].addClassName("active");
        activeTabId = tabId;
    };

    return tabStrip;

    function tabClick(tabId) {
        return function () {
            tabStrip.triggerCustomEvent("tabClicked", tabId);
        };
    }
};

//
// Widjet.extras.TabWidget
// TabWidget is a traditional tab control similar to what desktop applications often use.
//
// Widget hierarchy:
// Panel("div")
//    TabStrip("tabStrip")
//    Panel("div") - one per tab added
//
// Markup used:
// <div class="the className you provide in the constructor">
//    <ul class="tabStrip">
//        <li class="active">The selected tab</li>
//        <li>Another tab, not selected</li>
//        <li>Yet another tab, also not selected</li>
//    </ul>
//    <div>
//        Content panel for: The selected tab
//    </div>
//    <div class="hidden">
//        Content panel for: Another tab, not selected
//    </div>
//    <div class="hidden">
//        Content panel for: Yet another tab, also not selected<
//    </div>
// </div>
//
// Events:
// - No events
//
Widjet.extras.TabWidget = function (className) {
    var tabs = {};

    var tabWidget = Widjet.ui.Panel("div", className);
    var tabStrip = Widjet.extras.TabStrip("tabStrip");

    tabStrip.bindCustomEvent("tabClicked", tabClick);

    tabWidget.addWidget(tabStrip);

    tabWidget.addTab = function (text, tabId) {
        var tab = tabStrip.addTab(text, tabId);
        var tabPanel = Widjet.ui.Panel("div", "tabContent hidden");

        tabs[tabId] = { tab:tab, tabPanel:tabPanel};

        tabWidget.addWidget(tabPanel);

        // Make the added tab the active tab
        tabWidget.activeTab(tabId);
    };

    tabWidget.removeTab = function (tabId) {
        // Remember which tab is currently selected
        var activeTabId = tabStrip.activeTab();

        // Remove the tab we were handed the tabId for
        tabStrip.removeTab(tabId);
        tabWidget.removeWidget(tabWidget.tabContentPanel(tabId));
        delete tabs[tabId];

        // Did we remove the active tab?
        if (activeTabId !== tabId)
            return;

        // Yes we did, set the first tab as active, if there are any tabs left
        var tabIds = Object.keys(tabs);
        if (tabIds.length > 0)
            tabWidget.activeTab(tabIds[0]);
    };

    tabWidget.activeTab = function (tabId) {
        var activeTabId = tabStrip.activeTab();

        if (tabId === undefined)
            return activeTabId;

        if (activeTabId !== "")
            tabWidget.tabContentPanel(activeTabId).addClassName("hidden");

        tabStrip.activeTab(tabId);
        tabWidget.tabContentPanel(tabId).removeClassName("hidden");
    };

    tabWidget.tabStripPanel = function (tabId) {
        return tabs[tabId].tab;
    };

    tabWidget.tabContentPanel = function (tabId) {
        return tabs[tabId].tabPanel;
    };

    return tabWidget;

    function tabClick(tabId) {
        tabWidget.activeTab(tabId);
    }
};


//
// Model-View-Presenter help classes and protocols
//
// Model
// -----
// - Model is a single object, no other requirements
//
// View
// ----
// - A View instance must have a methods setData(obj).
//   The Presenter will call view.setData(model)
//   when the model has been updated for some reason.
//
// - Also, the responsibility of a View is to translate primitive
//   events like click, change etc and offer higher-level and more
//   application logic oriented events for the presenter to act upon.
//
// Presenter
// ---------
// - Sub-classes of the Presenter class needs to provide implementations
//   of addEventsToView(view) and removeEventsFromView(view)
//
// - These two methods should be implemented by sub-classes
//   and hook up the higher-level events from the views and update
//   the model accordingly
//
Widjet.mvp.Presenter = function () {
    var model = undefined;
    var views = [];

    var presenter = {};

    presenter.addView = function (view) {
        // Only add same view once
        var alreadyAdded = Widjet.util.any(views, function (v) {
            return v === view;
        });

        if (alreadyAdded)
            return;

        presenter.addEventsToView(view);

        views.push(view);
        presenter.updateView(view);
    };

    presenter.removeView = function (view) {
        for (var i = 0; i < views.length; i++) {
            if (views[i] == view) {
                presenter.removeEventsFromView(view);
                views.splice(i, 1);
                return;
            }
        }
    };

    presenter.updateViews = function () {
        Widjet.util.forEach(views, presenter.updateView);
    };

    presenter.updateView = function (view) {
        view.setData(model);
    };

    presenter.attachModel = function (newModel) {
        model = newModel;
        presenter.updateViews();
    };

    presenter.addEventsToView = function (view) {
        // Sub-classes should implement this
    };

    presenter.removeEventsFromView = function (view) {
        // Sub-classes should implement this
    };

    return presenter;
};
